import torch
from transformers import pipeline, AutomaticSpeechRecognitionPipeline, AutoModelForSpeechSeq2Seq, AutoProcessor
from typing import List, Tuple, Any, Dict
from pydantic import BaseModel, Field



class Chunk(BaseModel):
    """
    Represents a chunk of text with associated timestamps.
    """
    timestamp: Tuple[float, float] = Field(..., description="The timestamp range for the chunk of text.")
    text: str = Field(..., description="The text content of the chunk.")

class Transcription(BaseModel):
    """
    Represents the transcription output, consisting of text and chunks.
    """
    text: str = Field(..., description="The entire text transcription.")
    chunks: List[Chunk] = Field(..., description="List of individual text chunks with timestamps.")

    @property
    def timestamps(self) -> List[Tuple[float, float]]:
        """
        Get the list of timestamps for each chunk.

        Returns:
            List[Tuple[float, float]]: List of timestamps for each chunk.
        """
        return [chunk.timestamp for chunk in self.chunks]

    @property
    def chunk_texts(self) -> List[str]:
        """
        Get the list of texts for each chunk.

        Returns:
            List[str]: List of texts for each chunk.
        """
        return [chunk.text for chunk in self.chunks]


class SpeechInference:
    """
    Class for transcribing speech using the Hugging Face Transformers library.
    """

    def __init__(self, model_name: str, huggingface_read_token: str) -> None:
        """
        Initialize the SpeechTranscriber instance.

        Args:
            model_name (str): The name of the Hugging Face model to use for transcription.
            huggingface_read_token (str): The Hugging Face API token for model access.
        """
        self.model_name = model_name
        self.huggingface_read_token = huggingface_read_token
    
    def pipe_initialization(self) -> AutomaticSpeechRecognitionPipeline:
        """
        Initialize the pipeline for speech transcription.

        Returns:
            Any: The initialized pipeline object.
        """
        if input is None:
            print("No audio file submitted! Please upload or record an audio file before submitting your request.")
        else:
            device = 0 if torch.cuda.is_available() else "cpu"
            torch_dtype = torch.float16 if torch.cuda.is_available() else torch.float32
        model = AutoModelForSpeechSeq2Seq.from_pretrained(
        self.model_name , torch_dtype=torch_dtype, low_cpu_mem_usage=True, use_safetensors=True
        )
        model.to(device)
        processor = AutoProcessor.from_pretrained(self.model_name)
        pipe = pipeline(
                        task="automatic-speech-recognition",
                        model=model,
                        tokenizer=processor.tokenizer,
                        feature_extractor=processor.feature_extractor,
                        token=self.huggingface_read_token,
                        device=device,
                        torch_dtype=torch_dtype
                        )
        return pipe
    
    def output(self, pipe: AutomaticSpeechRecognitionPipeline, input: Any, task: str) -> Dict:
        """
        Perform speech transcription.

        Args:
            pipe (Any): The initialized pipeline object.
            input (Any): The input data to transcribe.
            task (str): The speech task, e.g., "transcribe" or "translate".

        Returns:
            Transcription: The transcription output.
        """
        transcription = pipe(
                input,
                chunk_length_s=30,
                batch_size=24,
                return_timestamps=True,
                generate_kwargs={"task": task}
            )
        # transcription = Transcription(**transcription)
        return transcription








import os
from transformers import Seq2SeqTrainingArguments, Seq2SeqTrainer
from .collator import DataCollatorSpeechSeq2SeqWithPadding
import evaluate
import torch
from datasets import DatasetDict
from .wandb_callback import WandbProgressResultsCallback
from transformers.trainer_pt_utils import IterableDatasetShard
from torch.utils.data import IterableDataset
from transformers import TrainerCallback
from .whisper_model_prep import WhisperModelPrep
from typing import Dict, Any
import warnings

warnings.filterwarnings("ignore")

class ShuffleCallback(TrainerCallback):
    def on_epoch_begin(self, args, state, control, train_dataloader, **kwargs):
        if isinstance(train_dataloader.dataset, IterableDatasetShard):
            pass
        elif isinstance(train_dataloader.dataset, IterableDataset):
            train_dataloader.dataset.set_epoch(train_dataloader.dataset._epoch + 1)


class Trainer:
    """

    A Trainer class for fine-tuning and training speech-to-text models using the Hugging Face Transformers library.

    """

    def __init__(
        self,
        huggingface_write_token: str,
        model_id: str,
        dataset: DatasetDict,
        model: str,
        feature_processor,
        feature_extractor,
        tokenizer,
        wandb_api_key: str,
        use_peft: bool,
    ):
        """
        Initializes the Trainer with the necessary components and configurations for training.

        Parameters:
            huggingface_push_token (str): Hugging Face API token for authenticated push access.
            model_id (str): Identifier for the pre-trained model.
            dataset (DatasetDict): The dataset split into 'train' and 'test'.
            model (PreTrainedModel): The model instance to be trained.
            feature_processor (Any): The audio feature processor.
            feature_extractor (Any): The audio feature extractor.
            tokenizer (PreTrainedTokenizer): The tokenizer for text data.
            language_abbr (str): Abbreviation for the dataset's language.
        """
        os.environ["WANDB_API_KEY"] = wandb_api_key
        self.dataset = dataset
        self.model = model
        self.model_id = model_id
        self.tokenizer = tokenizer
        self.feature_processor = feature_processor
        self.feature_extractor = feature_extractor
        self.huggingface_write_token = huggingface_write_token
        self.use_peft = use_peft
        self.model_prep = WhisperModelPrep(
            self.model_id, 
            "transcribe", 
            self.use_peft
        )

    def compute_metrics(self, pred) -> dict:
        """
        Computes the Word Error Rate (WER) metric for the model predictions.

        Parameters:
            pred (PredictionOutput): The output from the model's prediction.

        Returns:
            dict: A dictionary containing the computed WER metric.
        """
        pred_ids = pred.predictions
        label_ids = pred.label_ids
        label_ids[label_ids == -100] = self.tokenizer.pad_token_id
        pred_str = self.tokenizer.batch_decode(
            pred_ids, skip_special_tokens=True, normalize=True
        )
        label_str = self.tokenizer.batch_decode(
            label_ids, skip_special_tokens=True, normalize=True
        )
        metric = evaluate.load("wer")
        wer = 100 * metric.compute(predictions=pred_str, references=label_str)
        return {"wer": wer}
    

    def compute_spectrograms(self, example: Dict[str, Any]) -> Dict[str, Any]:
        """
        Computes spectrograms from audio waveform.

        Args:
            example (dict): A dictionary containing audio waveform data.

        Returns:
            dict: A dictionary containing computed spectrogram data.

        Raises:
            KeyError: If the input example does not contain the required keys.
        """
        waveform = example["audio"]["array"]
        feature_extractor = self.model_prep.initialize_feature_extractor()
        specs = feature_extractor(
            waveform, sampling_rate=16000, padding="do_not_pad"
        ).input_features[0]
        return {"spectrogram": specs}

    def train(self,
          max_steps: int = 100,
          learning_rate: float = 1e-5,
          per_device_train_batch_size: int = 96,
          per_device_eval_batch_size: int = 64,
          optim: str = "adamw_bnb_8bit"):
        """
        Conducts the training process using the specified model, dataset, and training configurations.

        Args:
            max_steps (int, optional): Maximum number of training steps. Defaults to 100.
            learning_rate (float, optional): Learning rate for training. Defaults to 1e-5.
            per_device_train_batch_size (int, optional): Batch size per GPU for training. Defaults to 96.
            per_device_eval_batch_size (int, optional): Batch size per GPU for evaluation. Defaults to 64.
            optim (str, optional): Optimizer to use for training. Defaults to "adamw_bnb_8bit".
        """
        # Checks if GPU is available: returns Boolean
        fp16 = torch.cuda.is_available()
        data_collator = DataCollatorSpeechSeq2SeqWithPadding(
            processor=self.feature_processor
        )
        training_args = Seq2SeqTrainingArguments(
            output_dir=f"../{self.model_id}-finetuned",
            per_device_train_batch_size=per_device_train_batch_size,
            gradient_accumulation_steps=1,
            learning_rate=learning_rate,
            max_steps=max_steps,
            gradient_checkpointing=True,
            fp16=fp16,
            optim=optim,
            evaluation_strategy="steps",
            per_device_eval_batch_size=per_device_eval_batch_size,
            predict_with_generate=True,
            generation_max_length=225,
            save_steps=25,
            eval_steps=25,
            logging_steps=25,
            load_best_model_at_end=True,
            metric_for_best_model="wer",
            greater_is_better=False,
            push_to_hub=True,
            hub_token=self.huggingface_write_token,
            report_to="wandb",
            remove_unused_columns=False,
            ignore_data_skip=True,
        )

        eval_dataset = self.dataset["test"].map(self.compute_spectrograms)

        trainer = Seq2SeqTrainer(
            args=training_args,
            model=self.model,
            train_dataset=self.dataset["train"],
            eval_dataset=eval_dataset,
            data_collator=data_collator,
            compute_metrics=self.compute_metrics,
            tokenizer=self.feature_processor.feature_extractor,
            callbacks=[ShuffleCallback()],
        )
        tokenizer = self.model_prep.initialize_tokenizer()
        processor = self.model_prep.initialize_processor()
        tokenizer.save_pretrained(training_args.output_dir)
        processor.save_pretrained(training_args.output_dir)
        torch.save(self.model.state_dict(), f"{training_args.output_dir}/pytorch_model.bin")
        self.model.save_pretrained(training_args.output_dir)
        progress_callback = WandbProgressResultsCallback(
            trainer, eval_dataset, tokenizer
        )
        trainer.add_callback(progress_callback)
        trainer.train()



import torch
import gradio as gr
from transformers import pipeline
import warnings
import yt_dlp as youtube_dl
import time
from transformers.pipelines.audio_utils import ffmpeg_read
import tempfile
import os
warnings.filterwarnings("ignore")

class WhisperDemo:
    def __init__(self, model_name, huggingface_read_token):
        self.model_name = model_name
        self.huggingface_read_token = huggingface_read_token
        self.pipe = None


    def initialize_pipeline(self):
        device = 0 if torch.cuda.is_available() else "cpu"
        if torch.cuda.is_available() and torch.cuda.is_bf16_supported():
            dtype = torch.bfloat16
        else:
            dtype = None

        self.pipe = pipeline(
            task="automatic-speech-recognition",
            model=self.model_name,
            token=self.huggingface_read_token,
            device=device,
            torch_dtype=dtype
        )


    def transcribe(self, inputs, task):
        if inputs is None:
            raise gr.Error("No audio file submitted! Please upload or record an audio file before submitting your request.")
        text = self.pipe(
                        inputs,
                        chunk_length_s=30,
                        batch_size=24, 
                        return_timestamps=True,
                        generate_kwargs={"task": task}
                        )["text"]
        return  text
    
    def _return_yt_html_embed(self, yt_url):
        video_id = yt_url.split("?v=")[-1]
        HTML_str = (
            f'<center> <iframe width="500" height="320" src="https://www.youtube.com/embed/{video_id}"> </iframe>'
            " </center>"
        )
        return HTML_str

    def download_yt_audio(self, yt_url, filename):
        YT_LENGTH_LIMIT_S = 3600
        info_loader = youtube_dl.YoutubeDL()
        
        try:
            info = info_loader.extract_info(yt_url, download=False)
        except youtube_dl.utils.DownloadError as err:
            raise gr.Error(str(err))
        
        file_length = info["duration_string"]
        file_h_m_s = file_length.split(":")
        file_h_m_s = [int(sub_length) for sub_length in file_h_m_s]
        
        if len(file_h_m_s) == 1:
            file_h_m_s.insert(0, 0)
        if len(file_h_m_s) == 2:
            file_h_m_s.insert(0, 0)
        file_length_s = file_h_m_s[0] * 3600 + file_h_m_s[1] * 60 + file_h_m_s[2]
        
        if file_length_s > YT_LENGTH_LIMIT_S:
            yt_length_limit_hms = time.strftime("%HH:%MM:%SS", time.gmtime(YT_LENGTH_LIMIT_S))
            file_length_hms = time.strftime("%HH:%MM:%SS", time.gmtime(file_length_s))
            raise gr.Error(f"Maximum YouTube length is {yt_length_limit_hms}, got {file_length_hms} YouTube video.")
        
        ydl_opts = {"outtmpl": filename, "format": "worstvideo[ext=mp4]+bestaudio[ext=m4a]/best[ext=mp4]/best"}
        
        with youtube_dl.YoutubeDL(ydl_opts) as ydl:
            try:
                ydl.download([yt_url])
            except youtube_dl.utils.ExtractorError as err:
                raise gr.Error(str(err))
            
    def yt_transcribe(self, yt_url, task, max_filesize=75.0):
        BATCH_SIZE = 8
        html_embed_str = self._return_yt_html_embed(yt_url)

        with tempfile.TemporaryDirectory() as tmpdirname:
            filepath = os.path.join(tmpdirname, "video.mp4")
            self.download_yt_audio(yt_url, filepath)
            with open(filepath, "rb") as f:
                inputs = f.read()

        inputs = ffmpeg_read(inputs, self.pipe.feature_extractor.sampling_rate)
        inputs = {"array": inputs, "sampling_rate": self.pipe.feature_extractor.sampling_rate}

        text = self.pipe(inputs, batch_size=BATCH_SIZE, generate_kwargs={"task": task}, return_timestamps=True)["text"]

        return html_embed_str, text

    def generate_demo(self):
        self.initialize_pipeline()

        mf_transcribe = gr.Interface(
            fn=self.transcribe,
            inputs=[
                gr.Audio(sources="microphone", type="filepath"),
                gr.Radio(["transcribe", "translate"], label="Task"),
            ],
            outputs="text",
            title="Transcribe Audio",
            description=(
                "Transcribe long-form microphone or audio inputs with the click of a button! Demo uses the OpenAI Whisper"
                f" checkpoint [{self.model_name}](https://huggingface.co/{self.model_name}) and 🤗 Transformers to transcribe audio files"
                " of arbitrary length."
            ),
            allow_flagging="never",
            )

        file_transcribe = gr.Interface(
            fn=self.transcribe,
            inputs=[
                gr.Audio(sources="upload", type="filepath", label="Audio file"),
                gr.Radio(["transcribe", "translate"], label="Task"),
            ],
            outputs="text",
            title="Transcribe Audio",
            description=(
                "Transcribe long-form microphone or audio inputs with the click of a button! Demo uses the OpenAI Whisper"
                f" checkpoint [{self.model_name}](https://huggingface.co/{self.model_name}) and 🤗 Transformers to transcribe audio files"
                " of arbitrary length."
            ),
            allow_flagging="never",
        )

        yt_transcribe = gr.Interface(
            fn=self.yt_transcribe,
            inputs=[
                gr.Textbox(lines=1, placeholder="Paste the URL to a YouTube video here", label="YouTube URL"),
                gr.Radio(["transcribe", "translate"], label="Task")
            ],
            outputs=["html", "text"],
            title="Transcribe YouTube",
            description=(
                "Transcribe long-form YouTube videos with the click of a button! Demo uses the OpenAI Whisper checkpoint"
                f" [{self.model_name}](https://huggingface.co/{self.model_name}) and 🤗 Transformers to transcribe video files of"
                " arbitrary length. The duration of this task will vary depending on the length of the video."
            ),
            allow_flagging="never",
            )

        demo = gr.TabbedInterface(
            [mf_transcribe, file_transcribe, yt_transcribe],
            ["Transcribe Audio", "Transcribe mp3 File", "Transcribe YouTube Video"],
        )
        demo.launch(share=True, debug=True)



from .gradio_inference import WhisperDemo
import argparse


def parse_args():
    parser = argparse.ArgumentParser(
        description="Whisper Demo: Transcribe Audio and YouTube"
    )
    parser.add_argument(
        "--model_name",
        type=str,
        help="Name of the fine-tuned model to use in your huggingfacehub repo",
    )

    parser.add_argument(
        "--huggingface_read_token",
        type=str,
        help="Hugging Face API token for read authenticated access.",
    )
    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()
    demo = WhisperDemo(
        args.model_name, args.huggingface_read_token
    )
    demo.generate_demo()



import os
from fastapi import FastAPI, UploadFile, HTTPException
from pydantic import BaseModel
import tempfile
from speech_inference import SpeechInference
import logging
import time
from dotenv import load_dotenv
load_dotenv()

app = FastAPI()


model_name = os.getenv("MODEL_NAME")
huggingface_read_token = os.getenv("HUGGINGFACE_READ_TOKEN")
inference = SpeechInference(model_name, huggingface_read_token)
pipeline = inference.pipe_initialization()


class AudioTranscriptionRequest(BaseModel):
    file: UploadFile
    task: str



@app.post("/speech_inference")
async def speechinference(file: UploadFile, task: str):
    logger = logging.getLogger(__name__)
    logging.basicConfig(format='%(asctime)s %(levelname)-8s %(message)s', filename='app.log', level=logging.INFO, datefmt='%Y-%m-%d %H:%M:%S')
    if file is None:
        logger.error("No file provided")
        raise HTTPException(status_code=400, detail="No file provided")

    try:
        with tempfile.NamedTemporaryFile(suffix=".tmp", delete=False) as tmp_file:
            tmp_file.write(await file.read())
            tmp_file_path = tmp_file.name

        if file.filename.endswith(".mp3"):
            logger.info("Processing mp3 file")
            start_time = time.time()
            result = inference.output(pipeline, tmp_file_path, task)
            end_time = time.time()
            logger.info(f"Time taken for inference: {end_time - start_time} seconds")
        elif file.filename.endswith(".wav"):
            logger.info("Processing wav file")
            pass
        else:
            logger.error("Unsupported file format")
            raise HTTPException(status_code=400, detail="Unsupported file format")

        logger.info("File processed successfully")
        return result
    except Exception as e:
        logger.error(f"Error processing file: {str(e)}")
        raise HTTPException(status_code=500, detail=f"Error processing file: {str(e)}")





import unittest
from unittest.mock import patch, MagicMock
import sys

sys.path.append("load_data")
from src.load_data import Dataset

# from load_data import Dataset


class TestDataset(unittest.TestCase):
    def setUp(self):
        self.huggingface_token = "your_token"
        self.dataset_name = "your_dataset_name"
        self.language_abbr = "your_language_abbr"
        self.dataset_manager = Dataset(
            self.huggingface_token, self.dataset_name, self.language_abbr
        )

    @patch("os.path.exists")
    @patch("datasets.load_dataset")
    def test_load_dataset_from_cache(self, mock_load_dataset, mock_os_path_exists):
        mock_os_path_exists.return_value = True
        # Mock the returned dataset
        mock_dataset = MagicMock()
        mock_load_dataset.return_value = mock_dataset

        dataset_dict = self.dataset_manager.load_dataset()

        mock_load_dataset.assert_called_with(
            self.dataset_name,
            self.language_abbr,
            split="train",
            token=self.huggingface_token,
            streaming=False,
            trust_remote_code=True,
            cache_dir=f"./{self.language_abbr}",
        )
        mock_load_dataset.assert_called_with(
            self.dataset_name,
            self.language_abbr,
            split="test",
            token=self.huggingface_token,
            streaming=False,
            trust_remote_code=True,
            cache_dir=f"./{self.language_abbr}",
        )
        self.assertEqual(dataset_dict["train"], mock_dataset)
        self.assertEqual(dataset_dict["test"], mock_dataset)

    @patch("os.path.exists")
    @patch("datasets.load_dataset")
    def test_load_dataset_from_remote(self, mock_load_dataset, mock_os_path_exists):
        mock_os_path_exists.return_value = False
        # Mock the returned dataset
        mock_dataset = MagicMock()
        mock_load_dataset.return_value = mock_dataset

        dataset_dict = self.dataset_manager.load_dataset()

        mock_load_dataset.assert_called_with(
            self.dataset_name,
            self.language_abbr,
            split="train",
            token=self.huggingface_token,
            streaming=False,
            trust_remote_code=True,
            cache_dir=f"./{self.language_abbr}",
        )
        mock_load_dataset.assert_called_with(
            self.dataset_name,
            self.language_abbr,
            split="test",
            token=self.huggingface_token,
            streaming=False,
            trust_remote_code=True,
            cache_dir=f"./{self.language_abbr}",
        )
        self.assertEqual(dataset_dict["train"], mock_dataset)
        self.assertEqual(dataset_dict["test"], mock_dataset)


if __name__ == "__main__":
    unittest.main()



import torch
from dataclasses import dataclass
from typing import Any, Dict, List, Union
import warnings

warnings.filterwarnings("ignore")

@dataclass
class DataCollatorSpeechSeq2SeqWithPadding:
    """
    A data collator that dynamically pads a batch of inputs for speech-to-text sequence-to-sequence models.

    Attributes:
        processor (Any): The processor combining feature extraction and tokenization,
                         used for preparing model inputs and labels.
    """

    processor: Any

    def __call__(
        self, features: List[Dict[str, Union[List[int], torch.Tensor]]]
    ) -> Dict[str, torch.Tensor]:
        """
        Processes and pads a batch of input features and labels for training or evaluation.

        Args:
            features (List[Dict[str, Union[List[int], torch.Tensor]]]): A list of feature dictionaries,
                each containing 'input_features' for model inputs and 'labels' for expected outputs.

        Returns:
            Dict[str, torch.Tensor]: A dictionary of padded tensors including 'input_features' for the model inputs
                                     and 'labels' for the model outputs, with padding applied according to the processor's
                                     configurations.
        """
        input_features = [
            {"input_features": feature["input_features"]} for feature in features
        ]
        batch = self.processor.feature_extractor.pad(
            input_features, return_tensors="pt"
        )

        label_features = [{"input_ids": feature["labels"]} for feature in features]
        labels_batch = self.processor.tokenizer.pad(label_features, return_tensors="pt")

        labels = labels_batch["input_ids"].masked_fill(
            labels_batch.attention_mask.ne(1), -100
        )

        if (labels[:, 0] == self.processor.tokenizer.bos_token_id).all().cpu().item():
            labels = labels[:, 1:]

        batch["labels"] = labels

        return batch



from transformers.integrations import WandbCallback
import panel as pn
import numpy as np
import wandb
import pandas as pd
import jiwer
import tempfile
import numbers
from pathlib import Path
from io import StringIO
import warnings
import holoviews as hv
hv.extension("bokeh", logo=False)
warnings.filterwarnings("ignore")


class RecordAnalyzer:
    """Class to analyze and process records from a dataset.
    """

    def __init__(self):
        self.tokenizer = None

    def record_to_html(self, sample_record) -> StringIO:
        """Convert a sample record to HTML format.

        Args:
        ----
            sample_record (dict): A sample record containing audio and spectrogram data.

        Returns:
        -------
            StringIO: HTML representation of the sample record.

        """
        audio_array = np.array(sample_record["audio"]["array"])
        audio_sr = sample_record["audio"]["sampling_rate"]
        audio_data = sample_record['audio']
        audio_duration = len(audio_data) / 16000
        audio_spectrogram = np.array(sample_record["spectrogram"])

        bounds = (0, 0, audio_duration, audio_spectrogram.max())

        waveform_int = np.int16(audio_array * 32767)

        hv_audio = pn.pane.Audio(
            waveform_int, sample_rate=audio_sr, name="Audio", throttle=500
        )

        slider = pn.widgets.FloatSlider(end=audio_duration, visible=False, step=0.001)
        line_audio = hv.VLine(0).opts(color="black")
        line_spec = hv.VLine(0).opts(color="red")

        slider.jslink(hv_audio, value="time", bidirectional=True)
        slider.jslink(line_audio, value="glyph.location")
        slider.jslink(line_spec, value="glyph.location")

        time = np.linspace(0, audio_duration, num=len(audio_array))
        line_plot_hv = (
            hv.Curve((time, audio_array), ["Time (s)", "amplitude"]).opts(
                width=500, height=150, axiswise=True
            )
            * line_audio
        )

        hv_spec_gram = (
            hv.Image(
                audio_spectrogram, bounds=(bounds), kdims=["Time (s)", "Frequency (hz)"]
            ).opts(width=500, height=150, labelled=[], axiswise=True, color_levels=512)
            * line_spec
        )

        combined = pn.Row(hv_audio, hv_spec_gram, line_plot_hv, slider)
        audio_html = StringIO()
        combined.save(audio_html)
        return audio_html

    def dataset_to_records(self, dataset) -> pd.DataFrame:
        """
        Convert a dataset to a DataFrame of records.

        Args:
            dataset (list): List of records in the dataset.

        Returns:
            pd.DataFrame: DataFrame containing the processed records.
        """
        records = []
        for item in dataset:
            record = {}
            audio_data = item['input_features']
            audio_duration = len(audio_data) / 16000
            record["audio_with_spec"] = wandb.Html(self.record_to_html(item))
            record["sentence"] = item["sentence"]
            record["length"] = audio_duration
            records.append(record)
        records = pd.DataFrame(records)
        return records

    def decode_predictions(self, predictions, tokenizer) -> list:
        """Decode model predictions into human-readable format.

        Args:
        ----
            predictions (object): Predictions generated by the model.

        Returns:
        -------
            list: Decoded predictions.

        """
        pred_ids = predictions.predictions
        pred_str = tokenizer.batch_decode(pred_ids, skip_special_tokens=True)
        return pred_str

    def compute_measures(self, predictions, labels) -> pd.DataFrame:
        """Compute evaluation measures for model predictions.

        Args:
        ----
            predictions (list): List of model predictions.
            labels (list): List of ground truth labels.

        Returns:
        -------
            pd.DataFrame: DataFrame containing computed evaluation measures.

        """
        measures = [
            jiwer.compute_measures(ls, ps) for ps, ls in zip(predictions, labels)
        ]
        measures_df = pd.DataFrame(measures)[
            ["wer", "hits", "substitutions", "deletions", "insertions"]
        ]
        return measures_df


class WandbProgressResultsCallback(WandbCallback):
    """Callback class for logging training progress to Weights & Biases.
    """

    def __init__(self, trainer, sample_dataset, tokenizer):
        """Initialize the WandbProgressResultsCallback instance.

        Args:
        ----
            trainer: The instance of the Trainer class.
            sample_dataset: The dataset used for logging sample predictions.
            tokenizer: The tokenizer instance for decoding predictions.

        """
        super().__init__()
        self.trainer = trainer
        self.sample_dataset = sample_dataset
        self.records_analyzer = RecordAnalyzer()
        self.records_df = self.records_analyzer.dataset_to_records(sample_dataset)
        self.tokenizer = tokenizer

    def on_log(self, args, state, control, model=None, logs=None, **kwargs):
        """Log training progress to Weights & Biases.

        Args:
        ----
            args: Arguments passed to the callback.
            state: The current state of the training.
            control: Control parameters for the callback.
            model: The trained model.
            logs: Additional logs.
            **kwargs: Additional keyword arguments.

        Returns:
        -------
            The table containing sample predictions.

        """
        super().on_log(args, state, control, model, logs)
        predictions = self.trainer.predict(self.sample_dataset)
        predictions = self.records_analyzer.decode_predictions(
            predictions, self.tokenizer
        )
        measures_df = self.records_analyzer.compute_measures(
            predictions, self.records_df["sentence"].tolist()
        )
        records_df = pd.concat([self.records_df, measures_df], axis=1)
        records_df["prediction"] = predictions
        records_df["step"] = state.global_step
        records_table = self._wandb.Table(dataframe=records_df)
        self._wandb.log({"sample_predictions": records_table})

    def on_save(self, args, state, control, model=None, tokenizer=None, **kwargs):
        """Save the trained model as an artifact in Weights & Biases.

        Args:
        ----
            args: Arguments passed to the callback.
            state: The current state of the training.
            control: Control parameters for the callback.
            model: The trained model.
            tokenizer: The tokenizer instance.
            **kwargs: Additional keyword arguments.

        """
        if self._wandb is None:
            return
        if self._log_model and self._initialized and state.is_world_process_zero:
            with tempfile.TemporaryDirectory() as temp_dir:
                self.trainer.save_model(temp_dir)
                metadata = (
                    {
                        k: v
                        for k, v in dict(self._wandb.summary).items()
                        if isinstance(v, numbers.Number) and not k.startswith("_")
                    }
                    if not args.load_best_model_at_end
                    else {
                        f"eval/{args.metric_for_best_model}": state.best_metric,
                        "train/total_floss": state.total_flos,
                    }
                )
                artifact = self._wandb.Artifact(
                    name=f"model-{self._wandb.run.id}", type="model", metadata=metadata
                )
                for f in Path(temp_dir).glob("*"):
                    if f.is_file():
                        with artifact.new_file(f.name, mode="wb") as fa:
                            fa.write(f.read_bytes())
                self._wandb.run.log_artifact(artifact)



from datasets import load_dataset, IterableDataset, concatenate_datasets
import warnings
from typing import List

warnings.filterwarnings("ignore")

class Dataset:
    """
    Manages loading of the datasets from Hugging Face's Common Voice repository.

    Attributes:
        huggingface_token (str): Hugging Face API token for read authenticated access.
        dataset_name (str): Name of the dataset to be downloaded from Hugging Face.
        language_abbr (str): Abbreviation of the language for the dataset.
    """

    def __init__(self, huggingface_token: str, dataset_name: str, language_abbr: List[str]):
        """
        Initializes the DatasetManager with necessary details for dataset operations.

        Parameters:
            huggingface_token (str): Hugging Face API token.
            dataset_name (str): Name of the dataset.
            language_abbr (str): Language abbreviation for the dataset.
        """
        self.huggingface_token = huggingface_token
        self.dataset_name = dataset_name
        self.language_abbr = language_abbr
    
    def load_dataset(self):
        """Load datasets for each language abbreviation and concatenate train/test splits.

        Returns:
            dict: A dictionary containing concatenated train and test splits for each language.
        """
        data = {}
        for lang in self.language_abbr:
            dataset = load_dataset(self.dataset_name, lang, streaming=True, token=self.huggingface_token, trust_remote_code=True, disable_progress_bar=True, disable_metadata=True)
            train_split = dataset['train']
            test_split = dataset['test']
            if "train" in data:
                data["train"] = concatenate_datasets([data["train"], train_split])
            else:
                data["train"] = train_split
            if "test" in data:
                data["test"] = concatenate_datasets([data["test"], test_split])
            else:
                data["test"] = test_split
        return data
        
    def count_examples(self, dataset: IterableDataset) -> int:
        """
        Count the number of examples in the dataset.

        Args:
            dataset (IterableDataset): The dataset to count examples from.

        Returns:
            int: The number of examples in the dataset.
        """
        count = 0
        for _ in dataset:
            count += 1
        return count


import argparse
from .data_prep import DataPrep
from .model_trainer import Trainer
import warnings

warnings.filterwarnings("ignore")

def parse_args():
    parser = argparse.ArgumentParser(
        description="Run the training orchestrator with specified parameters."
    )
    parser.add_argument(
        "--huggingface_read_token",
        type=str,
        required=True,
        help="Hugging Face API token for read authenticated access.",
    )
    parser.add_argument(
        "--huggingface_write_token",
        type=str,
        required=True,
        help="Hugging Face API token for write authenticated access.",
    )
    parser.add_argument(
        "--dataset_name",
        type=str,
        default="mozilla-foundation/common_voice_16_1",
        help="Name of the dataset to be downloaded from Hugging Face.",
    )
    parser.add_argument(
        "--language_abbr",
        nargs='+',  
        required=True,
        help="Abbreviation(s) of the language(s) for the dataset.",
    )
    parser.add_argument(
        "--model_id",
        type=str,
        default="openai/whisper-small",
        help="Model ID for the model to be used in training.",
    )
    parser.add_argument(
        "--processing_task",
        type=str,
        default="transcribe",
        help="The processing task to be performed.",
    )
    parser.add_argument(
        "--wandb_api_key",
        type=str,
        help="The wandb.ai api key for monitoring training runs, signup and generate an api key.",
    )
    parser.add_argument(
        "--use_peft",
        type=bool,
        help="True to train your model using PEFT method, False for full finetuning",
    )

    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()
    process = DataPrep(
        huggingface_read_token=args.huggingface_read_token,
        dataset_name=args.dataset_name,
        language_abbr=args.language_abbr,
        model_id=args.model_id,
        processing_task=args.processing_task,
        use_peft=args.use_peft,
    )
    tokenizer, feature_extractor, feature_processor, model = process.prepare_model()

    dataset = process.load_dataset(feature_extractor, tokenizer, feature_processor)
    trainer = Trainer(
        huggingface_write_token=args.huggingface_write_token,
        model_id=args.model_id,
        dataset=dataset,
        model=model,
        feature_processor=feature_processor,
        feature_extractor=feature_extractor,
        tokenizer=tokenizer,
        wandb_api_key=args.wandb_api_key,
        use_peft=args.use_peft,
    )
    trainer.train()
